from __future__ import annotations

# Import builtin recipes for side effects (registration).
from .industrial_adapt import industrial_adapt  # noqa: F401
from .anomalib_train import anomalib_train  # noqa: F401
from .micro_finetune_autoencoder import micro_finetune_autoencoder  # noqa: F401
